from ExKMC.Tree import Tree, Node
import numpy as np
from sklearn.cluster import KMeans
import numpy


class RandomThresholdTree(Tree):
    """
    Tree constructor for explainable k-means clustering
    using MAkarychev's algorithm.
    :param k: Number of clusters.
    :param depth_factor: Weight of penalty term to disincentivize deep trees.
    :param random_state: Random seed generator for kmeans.
    """

    def __init__(self, k, random_state=0):
        super().__init__(k, random_state=random_state)
        self.base_tree = 'RandomThreshold'
        self.rdm = np.random.default_rng(seed=int(self.random_state))

    def fit(self, x_data, centers=None):
        kmeans = KMeans(self.k, verbose=self.verbose, 
                        random_state=self.random_state)
        if centers is None:
            if self.verbose > 0:
                print('Finding %d-means' % self.k)
            kmeans.fit(x_data)
        else:
            kmeans.cluster_centers_ = centers
        
        y = np.array(kmeans.predict(x_data), dtype=np.int32)
        centers = np.array(kmeans.cluster_centers_, dtype=np.float64)
        self.tree = self._fit_tree(x_data, centers)
        self._feature_importance = np.zeros(x_data.shape[1])
        self.__fill_stats__(self.tree, x_data, y)

        return self
    
    def _fit_tree(self, data, centers):
        """
        Calculates the distances between all data and all centers from an
        unrestricted partition and finds a tree that induces an explainable
        partition based on the unrestricted one.
        """
        k, _ = centers.shape
        unique_data, data_count = np.unique(data, axis=0, return_counts=True)
        n = unique_data.shape[0]
        valid_centers = np.ones(k, dtype=bool)
        valid_data = np.ones(n, dtype=bool)
        phi_data, phi_centers = embedding(unique_data,centers)
        return self._build_tree(unique_data, data_count, centers,  0,
                                valid_centers, valid_data,phi_data,phi_centers)
        
    def _build_tree(self, data, data_count, centers, cur_height,
                    valid_centers, valid_data, phi_data, phi_centers):
        """
        Builds a tree that induces an explainable partition (from axis-aligned
        cuts) of the data, based on the centers provided by an unrestricted
        partition.
        """
        node = Node()
        k = valid_centers.sum()
        n = valid_data.sum()
        if k == 1:
            node.value = np.argmax(valid_centers)
            return node

        curr_data = data[valid_data]
        curr_centers = centers[valid_centers]
        curr_phi_data = phi_data[valid_data]
        curr_phi_centers = phi_centers[valid_centers]
        dim, cut,terminal = self._best_cut(data, data_count, valid_data,
                                valid_centers,phi_data, phi_centers)
        #here the dim and cut are in embedded space, so we need to find the corresponding cut in the original space

        if terminal:
            node.value = np.argmax(valid_centers)
            return node

        highest_center_value_below = -np.inf 
        highest_data_value_below = -np.inf 
        smallest_data_value_over = np.inf 
        smallest_center_value_over = np.inf
        highest_center_value_below_idx = -1 
        highest_data_value_below_idx = -1 
        smallest_data_value_over_idx = -1 
        smallest_center_value_over_idx = -1
        
        # the embedding function is increasing over R. 
        # we find which is the largest value below the cutoff and which is the smallest above 
        # and define the cutoff in original space to be the midpoint of the two; 
        # i.e., the cut that represents an equivalent result in terms of separation
        for i in range(k):
            if(curr_phi_centers[i,dim]<=cut): #below
                if(curr_phi_centers[i,dim]>highest_center_value_below):
                    highest_center_value_below = curr_phi_centers[i,dim]
                    highest_center_value_below_idx = i
            else: #over
                if(curr_phi_centers[i,dim] < smallest_center_value_over):
                    smallest_center_value_over = curr_phi_centers[i,dim] 
                    smallest_center_value_over_idx = i

        for i in range(n):
            if(curr_phi_data[i,dim]<=cut): #below
                if(curr_phi_data[i,dim]>highest_data_value_below):
                    highest_data_value_below = curr_phi_data[i,dim]
                    highest_data_value_below_idx = i
            else: #over
                if(curr_phi_data[i,dim] < smallest_data_value_over):
                    smallest_data_value_over = curr_phi_data[i,dim] 
                    smallest_data_value_over_idx = i

        if(highest_data_value_below > highest_center_value_below):
            original_highest_below = curr_data[highest_data_value_below_idx,dim]
        else:
            original_highest_below = curr_centers[highest_center_value_below_idx,dim]

        if(smallest_center_value_over > smallest_data_value_over):
            original_smallest_over = curr_data[smallest_data_value_over_idx,dim]
        else:
            original_smallest_over = curr_centers[smallest_center_value_over_idx,dim]

        original_cut = (original_smallest_over+original_highest_below)/2


        node.feature = dim
        node.value = original_cut

        n = data.shape[0]
        data_below = 0
        left_valid_data = np.zeros(n, dtype=bool)
        right_valid_data = np.zeros(n, dtype=bool)
        for i in range(n):
            if valid_data[i]:
                if data[i,dim] <= original_cut:
                    left_valid_data[i] = True
                    data_below += 1
                else:
                    right_valid_data[i] = True

        k = centers.shape[0]
        centers_below = 0
        left_valid_centers = np.zeros(k, dtype=bool)
        right_valid_centers = np.zeros(k, dtype=bool)
        for i in range(k):
            if valid_centers[i]:
                if centers[i, dim] <= original_cut:
                    left_valid_centers[i] = True
                    centers_below += 1
                else:
                    right_valid_centers[i] = True

        node.left = self._build_tree(data, data_count, centers, cur_height + 1,
                                left_valid_centers, left_valid_data,phi_data,phi_centers)
        node.right = self._build_tree(data, data_count, centers,  cur_height + 1,
                                right_valid_centers, right_valid_data,phi_data,phi_centers)
        return node
   
    def _best_cut(self, data, data_count, valid_data, valid_centers, phi_data, phi_centers):
        """
        Finds the best cut across any dimension of data.
        """
        best_cut = -np.inf
        best_dim = -1

        n = valid_data.sum()
        k = valid_centers.sum()

        terminal = False
        
        ans = self._get_best_cut(data, data_count, valid_data, valid_centers,
                                    n, k,phi_data, phi_centers)
        best_dim, best_cut = ans
        if best_cut == -np.inf:
            terminal = True
        return best_dim, best_cut, terminal

    def _get_best_cut(self, data, data_count, valid_data, valid_centers,
                        n, k,phi_data, phi_centers):
        dim = len(data[0])
        
        phi_data = phi_data[valid_data]
        data_count = data_count[valid_data]

        phi_centers = phi_centers[valid_centers]

        ##### ALGORITHM
        # unseparated centers are sorted for each dimension
        # given the order, the union of cuts that separate centers is [c1,cn[
        # let c_{ij} be the j-th center (in order) in dimension i
        # At will be [[1,c_{11},c_{1m}],[2,c_{21,c2m}],....,[d,c_{d1},c_{dm}]], 
        # where m is the number of unseparated centers
        At = []
        
        for i in range(dim):
            # cut is possible if separates at least 2 centers
            # (i.e., there is a center to the right of it)
            # cut with no center to the right is last_center
            At.append([i])
            phi_centers_dim = phi_centers[:,i]
            phi_centers_dim_sort = np.argsort(phi_centers_dim)
            last_phi_center = phi_centers_dim[phi_centers_dim_sort[-1]]
            first_phi_center = phi_centers_dim[phi_centers_dim_sort[0]]
            if(last_phi_center > first_phi_center):
                At[-1].append(first_phi_center)
                At[-1].append(last_phi_center)
        total_length =0
        for i in range(dim):
            if(len(At[i])==3):
                total_length += At[i][2] - At[i][1]

        auxiliar_length = self.rdm.uniform(0,total_length)
        best_dim = -1
        best_cut = -1
        for i in range(dim):
            if(len(At[i])==3):
                auxiliar_length = auxiliar_length -(At[i][2] - At[i][1])
                if(auxiliar_length<0):
                    auxiliar_length+=At[i][2] - At[i][1]
                    best_cut = At[i][1] + auxiliar_length
                    best_dim = At[i][0]
                    break

        if(best_dim ==-1):
            #in which case the draw gives total_length. 
            #As the interval is open, I define that it will be the same as when the draw gives 0. 
            #This happens with probability 0
            for i in range(dim):
                if(len(At[i])==3):
                    best_dim = At[0]
                    best_cut = At[1]
        return best_dim,best_cut

def embedding(data,centers):
    #data = data points d-dimensional
    #center = centers points d-dimensional
    valid_k = len(centers)
    valid_n = len(data)
    dim = len(data[0])
    phi_n = np.zeros((valid_n,dim))
    phi_k = np.zeros((valid_k,dim))
    for i in range(dim):
        centers_dim = centers[:,i]
        arg_sort_center = np.argsort(centers_dim)
        data_dim = data[:,i]
        arg_sort_data = np.argsort(data_dim)
        aux_n,aux_k = embedding_dim(data_dim,centers_dim, arg_sort_data, arg_sort_center)
        phi_n[:,i] = aux_n 
        phi_k[:,i] = aux_k 
    return phi_n,phi_k


def embedding_dim (data_dim,center_dim, arg_sort_data, arg_sort_center):
    #data_dim = projection of data in a specific dimension
    #center_dim = projection of centes in a specific dimension
    #arg_sort_data = pointers to ordered positions in data vector
    #ard_sort_center = pointers to orderes positions in center vector
    valid_k = len(center_dim)
    valid_n = len(data_dim)
    phi_k =  np.zeros(valid_k)
    phi_n = np.zeros(valid_n)
    for i in range(1,valid_k):
        phi_k[arg_sort_center[i]] = phi_k[arg_sort_center[i-1]]+((center_dim[arg_sort_center[i]]-center_dim[arg_sort_center[i-1]])**2)/2
    for i in range(valid_n):
        idx = -1
        small_dist = np.inf 
        for j in range(valid_k):
            curr_dist = abs(center_dim[arg_sort_center[j]]-data_dim[arg_sort_data[i]])
            if(curr_dist<small_dist):
                idx = j
                small_dist = curr_dist

        sgn = signal(data_dim[arg_sort_data[i]] - center_dim[arg_sort_center[idx]])
        phi_n[arg_sort_data[i]] = phi_k[arg_sort_center[idx]] + sgn*(data_dim[arg_sort_data[i]]-center_dim[arg_sort_center[idx]])**2
    return phi_n,phi_k 


def signal(x):
    if(x>0):
        return 1
    elif(x<0):
        return -1
    else:
        return 0


from ExKMC.Tree import Tree, Node
import numpy as np

class BisectionTree(Tree):
    """
    Explainable k-means clustering without the need of a previous partition of the data.
    :param k: Number of clusters.
    """

    def __init__(self, k):
        super().__init__(k)
        self.base_tree = 'Bisection'
        self.k = k
    
    def fit(self, data):
        cuts = self.get_cuts(data)
        self.tree = self.fill_tree(cuts)
    
    def get_best_cut(self, data):
        order = np.argsort(data, 0)
        _, d = data.shape
        best_gain = best_dim = best_val = -np.inf
        for i in range(d):
            val, gain = self.get_best_for_dim(data, order[:,i], i)
            # print(i, val, gain)
            if gain > best_gain:
                best_dim = i
                best_val = val
                best_gain = gain
        return best_dim, best_val, best_gain

    def get_best_for_dim(self, data, order, dim):
        vals, indices = np.unique(data[order, dim], return_index=True)
        if len(vals) <= 1: # cannot separate data on this dimension
            return -np.inf, -np.inf
        indices = indices[1:] - 1
        counts = indices + 1
        cs = data[order].cumsum(0)[indices]
        css = np.square(cs)
        cssm = css / counts[:,None]
        rev_cs = data.sum(0) - cs
        rev_css = np.square(rev_cs)
        rev_counts = len(data) - counts
        rev_cssm = rev_css / rev_counts[:,None]
        gains = (cssm + rev_cssm).sum(1)
        idx = gains.argmax()
        val = vals[idx]
        return val, gains[idx]

    def get_cuts(self, data):
        n, _ = data.shape
        vars_ = np.zeros(self.k)
        mask = np.zeros(n, dtype=int)
        check = np.zeros(self.k, dtype=int)
        check[0] = 1
        lengths = np.zeros(self.k)
        lengths[0] = len(data)
        cuts = []
        n_clusters = 1
        vars_[0] = data.var(axis=0).sum() * len(data)
        while n_clusters < self.k:
            # print('find cluster', n_clusters)
            best_k = best_val = best_gain = best_dim = -np.inf
            best_var0 = best_var1 = -np.inf
            for i in range(n_clusters):
                if check[i]:
                    # print('cluster', i)
                    cur_data = data[mask==i]
                    dim, val, _ = self.get_best_cut(cur_data)
                    d0 = data[(mask==i) & (data[:,dim] <= val)]
                    var0 = d0.var(0).sum() * len(d0)
                    d1 = data[(mask==i) & (data[:,dim] > val)]
                    var1 = d1.var(0).sum() * len(d1)
                    gain = vars_[i] - var0 - var1
                    # print(gain, vars_[i], var0, var1)
                    if gain > best_gain:
                        best_gain = gain
                        best_k = i
                        best_dim = dim
                        best_val = val
                        best_var0 = var0
                        best_var1 = var1
                    # print()
            # print(best_k, best_dim, best_val)
            cuts.append((best_k, best_dim, best_val))
            new_mask = (mask == best_k) & (data[:,best_dim] > best_val)
            mask[new_mask] = n_clusters
            # print(mask)
            vars_[best_k] = best_var0
            vars_[n_clusters] = best_var1
            check[n_clusters] = 1
            n_clusters += 1
        return cuts

    def fill_tree(self, cuts):
        node = Node()
        node.samples = 0
        for i in range(len(cuts)):
            c, dim, cut = cuts[i]
            to_split = self.find_node(node, c)
            if to_split is None:
                print(f'node {c} not found')
            to_split.feature = dim
            to_split.value = cut
            to_split.left = Node()
            to_split.left.samples = c
            to_split.right = Node()
            to_split.right.samples = i+1
            to_split.samples = None
        return self.fix_tree(node)

    def find_node(self, node, c):
        if node.samples == c:
            return node
        if node.left:
            left = self.find_node(node.left, c)
            if isinstance(left, Node):
                return left
        if node.right:
            right = self.find_node(node.right, c) 
            if isinstance(right, Node):
                return right
        return None
    
    def fix_tree(self, node):
        nodes = [node]
        while nodes:
            n = nodes.pop()
            if n.left:
                nodes.append(n.left)
            if n.right:
                nodes.append(n.right)
            if n.is_leaf():
                n.value = n.samples
        return node


from ExKMC.Tree import Tree, Node
import numpy as np
from sklearn.cluster import KMeans
import ctypes as ct
import pathlib

libfile = pathlib.Path(__file__).parent / "lib_best_cut.so"
LIB = ct.CDLL(libfile)
    
C_FLOAT_P = ct.POINTER(ct.c_float)
C_INT_P = ct.POINTER(ct.c_int)

LIB.best_cut_single_dim.restype = ct.c_void_p
LIB.best_cut_single_dim.argtypes = [C_FLOAT_P, C_INT_P, C_FLOAT_P, 
                                     C_FLOAT_P, C_INT_P, ct.c_int, 
                                     ct.c_int, C_FLOAT_P, ct.c_double, 
                                     ct.c_bool, ct.c_bool]

class ShallowTree(Tree):
    """
    Shallow tree constructor for explainable k-means clustering.
    :param k: Number of clusters.
    :param depth_factor: Weight of penalty term to disincentivize deep trees.
    :param random_state: Random seed generator for kmeans.
    """

    def __init__(self, k, depth_factor=0.03, random_state=None):
        super().__init__(k, random_state=random_state)
        self.depth_factor = depth_factor
        self.base_tree = 'Shallow' if depth_factor else 'ExGreedy'

    def fit(self, x_data, centers=None):
        kmeans = KMeans(self.k, verbose=self.verbose, 
                        random_state=self.random_state)
        if centers is None:
            if self.verbose > 0:
                print('Finding %d-means' % self.k)
            kmeans.fit(x_data)
        else:
            kmeans.cluster_centers_ = centers
        
        y = np.array(kmeans.predict(x_data), dtype=np.int32)
        centers = np.array(kmeans.cluster_centers_, dtype=np.float64)
        self.tree = self._fit_tree(x_data, centers, self.depth_factor)
        self._feature_importance = np.zeros(x_data.shape[1])
        self.__fill_stats__(self.tree, x_data, y)

        return self
    
    def _fit_tree(self, data, centers, depth_factor):
        """
        Calculates the distances between all data and all centers from an
        unrestricted partition and finds a tree that induces an explainable
        partition based on the unrestricted one.
        """
        k, d = centers.shape
        unique_data, data_count = np.unique(data, axis=0, 
                                            return_counts=True)
        n = unique_data.shape[0]
        valid_centers = np.ones(k, dtype=bool)
        valid_data = np.ones(n, dtype=bool)
        distances = get_distances(unique_data, centers)
        cuts_matrix = np.zeros((d,2), dtype=int)
        return self._build_tree(unique_data, data_count, centers,
                                distances, valid_centers, valid_data,
                                depth_factor, cuts_matrix)
    
    def _build_tree(self, data, data_count, centers,
                    distances, valid_centers, valid_data,
                    depth_factor, cuts_matrix):
        """
        Builds a tree that induces an explainable partition (from 
        axis-aligned cuts) of the data, based on the centers provided 
        by an unrestricted partition.
        """
        node = Node()
        k = valid_centers.sum()
        n = valid_data.sum()
        if k == 1:
            node.value = np.argmax(valid_centers)
            return node

        dim, cut, _, terminal = self._best_cut(data, data_count, valid_data, 
                                         centers, valid_centers, distances, 
                                         depth_factor, cuts_matrix)
        if terminal:
            node.value = np.argmax(valid_centers)
            return node

        node.feature = dim
        node.value = cut

        n = data.shape[0]
        data_below = 0
        left_valid_data = np.zeros(n, dtype=bool)
        right_valid_data = np.zeros(n, dtype=bool)
        for i in range(n):
            if valid_data[i]:
                if data[i,dim] <= cut:
                    left_valid_data[i] = True
                    data_below += 1
                else:
                    right_valid_data[i] = True

        k = centers.shape[0]
        centers_below = 0
        left_valid_centers = np.zeros(k, dtype=bool)
        right_valid_centers = np.zeros(k, dtype=bool)
        for i in range(k):
            if valid_centers[i]:
                if centers[i, dim] <= cut:
                    left_valid_centers[i] = True
                    centers_below += 1
                else:
                    right_valid_centers[i] = True

        cuts_matrix[node.feature,0] += 1
        node.left = self._build_tree(data, data_count, centers,
                                     distances, left_valid_centers, 
                                     left_valid_data, depth_factor, 
                                     cuts_matrix)
        cuts_matrix[node.feature,0] -= 1
        cuts_matrix[node.feature,1] += 1
        node.right = self._build_tree(data, data_count, centers,
                                      distances, right_valid_centers, 
                                      right_valid_data, depth_factor, 
                                      cuts_matrix)
        cuts_matrix[node.feature,1] -= 1
        return node
    
    def _best_cut(self, data, data_count, valid_data, centers, 
                    valid_centers, distances, depth_factor, cuts_matrix):
        """
        Finds the best cut across any dimension of data.
        """
        dim = centers.shape[1]
        best_cut = -np.inf
        best_dim = -1
        best_cost = np.inf

        n = valid_data.sum()
        k = valid_centers.sum()

        terminal = False

        for i in range(dim):
            if len(np.unique(data[valid_data,i])) == 1:
                continue
            ans = self._get_best_cut_dim(data, data_count, valid_data, 
                                        centers, valid_centers, distances, 
                                        n, k, i, LIB.best_cut_single_dim, 
                                        depth_factor, cuts_matrix[i])
            cut, cost = ans
            if cost < best_cost:
                best_cut = cut
                best_dim = i
                best_cost = cost
        if best_cut == -np.inf:
            terminal = True
        return best_dim, best_cut, best_cost, terminal
    
    def _get_best_cut_dim(self, data, data_count, valid_data, centers, 
                            valid_centers, distances, n, k, dim, func, 
                            depth_factor, cuts_row):
        """
        Calls the C function that finds the cut in data (across dimension 
        dim) with the smallest cost.
        """

        full_dist_mask = np.outer(valid_data, valid_centers)
        distances_f = np.asarray(distances[full_dist_mask], 
                                    dtype=np.float64)
        distances_pointer = distances_f.ctypes.data_as(C_FLOAT_P)

        dist_shape = distances_f.reshape(n, k)
        dist_order = np.argsort(dist_shape, axis=1)
        dist_order_f = np.asarray(dist_order, 
                                    dtype=np.int32).reshape(n*k)
        dist_order_pointer = dist_order_f.ctypes.data_as(C_INT_P)

        data_f = np.asarray(data[valid_data, dim], 
                            dtype=np.float64)
        data_p = data_f.ctypes.data_as(C_FLOAT_P)

        data_count_f = np.asarray(data_count[valid_data], 
                                    dtype=np.int32)
        data_count_p = data_count_f.ctypes.data_as(C_INT_P)

        centers_f = np.asarray(centers[valid_centers,dim], 
                                dtype=np.float64)
        centers_p = centers_f.ctypes.data_as(C_FLOAT_P)

        bool_cut_left = bool(cuts_row[0])
        bool_cut_right = bool(cuts_row[1])

        ans = np.zeros(2, dtype=np.float64)
        ans_p = ans.ctypes.data_as(C_FLOAT_P)
        func(data_p, data_count_p, centers_p, distances_pointer,
            dist_order_pointer, n, k, ans_p, depth_factor, 
            bool_cut_left, bool_cut_right)
        return ans

def get_distances(data, centers):
    """
    Finds the squared Euclidean distances between each data point in 
    data and each center in centers.
    """
    distances = np.zeros((data.shape[0], centers.shape[0]))
    for i in range(centers.shape[0]):
        distances[:,i] = np.linalg.norm(data - centers[i], axis=1) ** 2
    return distances

# if __name__ == '__main__':
#     import joblib, sys
#     from sklearn.cluster import KMeans
#     data_name = sys.argv[1]
#     seed = int(sys.argv[2]) if len(sys.argv) >= 3 else None
#     folder = '/home/lmurtinho_local/shallow_decision_trees/results/data'
#     data_dict = joblib.load(f'{folder}/{data_name}.joblib')
#     data = data_dict['data']
#     k = data_dict['k']
#     km = KMeans(k, random_state=seed)
#     km.fit(data)
#     st = ShallowTree(k=k)
#     st.fit(data, km)
#     print(st.score(data))
#     t = Tree(k)
#     t.fit(data, km)
#     print(t.score(data))


from setuptools import setup, Extension
from ShallowTree import __version__ as version
with open('README.md', 'r') as f:
    long_description = f.read()

#  https://stackoverflow.com/questions/4529555/building-a-ctypes-based-c-library-with-distutils
from distutils.command.build_ext import build_ext as build_ext_orig
class CTypesExtension(Extension): pass
class build_ext(build_ext_orig):

    def build_extension(self, ext):
        self._ctypes = isinstance(ext, CTypesExtension)
        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        if self._ctypes:
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        if self._ctypes:
            return ext_name + '.so'
        return super().get_ext_filename(ext_name)


setup(
    name="ShallowTree",
    version=version,
    author="Lucas Murtinho",
    author_email="lucas.murtinho@gmail.com",
    license="MIT",
    description="Shallow decision trees for explainable clustering",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lmurtinho/ShallowTree",
    py_modules=['ShallowTree','BisectionTree','RandomThresholdTree'],
    ext_modules=[CTypesExtension('ShallowTree/lib_best_cut', 
                                ['ShallowTree/best_cut.c'])],
    install_requires=['numpy', 'scikit-learn', 'ExKMC'],
    cmdclass={'build_ext': build_ext},
)